/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <libintl.h>

#include "sambascanner.h"
#include "Tools.h"
#include "IP_Range.h"
#include "Processes.h"
#include "String.h"
#include "Parser.h"
#include "Mutex.h"
#include "Threads.h"

extern struct Options_Structure Options;

int GetUniqueThreadID() {
	static pthread_mutex_t LockUniqueID = PTHREAD_MUTEX_INITIALIZER;
	static int UniqueID = 0;
	int ReturnID;
	pthread_mutex_lock(&LockUniqueID);
	UniqueID++;
	ReturnID = UniqueID;
	pthread_mutex_unlock(&LockUniqueID);
	return ReturnID;
}

void* Redirect_Thread(void *Parameters) {
	struct Redirect_Thread_InputData *Input = (struct Redirect_Thread_InputData*)Parameters;
	char Buffer[BUFSIZE_SHORT];
	int Read;
	int Cnt = 0;
	int First = 1;
	int TotalRead = 0;
	int TotalWritten = 0;

	pthread_mutex_lock(&Input->Active);

	do {
		Cnt++;
		int Error;
		errno = 0;
		Read = read(Input->Read_FD, Buffer, sizeof(Buffer));
		Error = errno;
		message(5, _("[UniqTID %d] Read #%d (after %d read, %d written): Read = %d, errno = %d (\"%s\") (ReadFD = %d -> OutfileFD = %d)\n"), Input->UniqueThreadID, Cnt, TotalRead, TotalWritten, Read, Error, strerror(Error), Input->Read_FD, Input->Outfile_FD);
		if (Read < 0) {
			message(4, _("[UniqTID %d] Invalid read #%d (after %d read, %d written): Read = %d, errno = %d (\"%s\") (ReadFD = %d -> OutfileFD = %d)\n"), Input->UniqueThreadID, Cnt, TotalRead, TotalWritten, Read, Error, strerror(Error), Input->Read_FD, Input->Outfile_FD);
		} else {
			TotalRead += Read;
		}
		/* TODO: This error sometimes (rarely) happens - but why? */
		/*
		if (Error == EBADF) {
			printf("ERROR DETECTED!\n");
			fflush(stdout);
			fflush(stderr);
			*((int*)0) = 1;
		}
		*/
		if ((First) && (Error == EBADF)) {
			/* Wait for FD to become valid */
			message(4, _("[UniqTID %d] Waiting for FD to become valid after EBADF.\n"), Input->UniqueThreadID);
			usleep(100 * 1000);
			continue;
		}
		if (Read > 0) {
			double Time;
			int Written;

			First = 0;
			errno = 0;
			Written = write(Input->Outfile_FD, Buffer, Read);
			Error = errno;
			message(5, _("[UniqTID %d] Write #%d (after %d read, %d written): Written = %d, errno = %d (\"%s\") (ReadFD = %d -> OutfileFD = %d)\n"), Input->UniqueThreadID, Cnt, TotalRead, TotalWritten, Read, Error, strerror(Error), Input->Read_FD, Input->Outfile_FD);
			if (Written < 0) {
				message(4, _("[UniqTID %d] Invalid write #%d (after %d read, %d written): Written = %d, errno = %d (\"%s\") (ReadFD = %d -> OutfileFD = %d)\n"), Input->UniqueThreadID, Cnt, TotalRead, TotalWritten, Read, Error, strerror(Error), Input->Read_FD, Input->Outfile_FD);
			} else {
				TotalWritten += Written;
			}
			Time = GetTime();
			(*Input->Processes)[Input->ProcessID].TermTime = Time + (Input->TermTime * Options.Timings);
			(*Input->Processes)[Input->ProcessID].KillTime = Time + ((Input->TermTime + Input->KillTime) * Options.Timings);
		}
	} while (Read > 0);
	message(4, _("[UniqTID %d] FD %d -> Redirect_Thread -> FD %d Quit.\n"), Input->UniqueThreadID, Input->Read_FD, Input->Outfile_FD);
	close(Input->Read_FD);
	close(Input->Outfile_FD);

	free(Input);		/* Invalid write! */
	pthread_exit(NULL);
	return NULL;
}

void* Process_Watcher_Thread(void *Parameters) {
	struct Process_Watcher_Thread_InputData *Input = (struct Process_Watcher_Thread_InputData*)Parameters;
	int i;
	double Time;

	while (Input->Running) {
		Time = GetTime();
		for (i = 0; i < *(Input->NumberProcesses); i++) {
			if (((*Input->Processes)[i].Running == 1) && ((*Input->Processes)[i].Valid == 1)) {
				if (Time > (*Input->Processes)[i].KillTime) {
					if ((*Input->Processes)[i].Signalling_State == 1) {
						message(3, _("=> SIGKILL Processes[%d] [PID %d], t = %.0f, KillTime = %.0f\n"), i, (*Input->Processes)[i].PID, Time, (*Input->Processes)[i].KillTime);
						(*Input->Processes)[i].Signal = SIGKILL;
						kill((*Input->Processes)[i].PID, SIGKILL);
						(*Input->Processes)[i].Signalling_State = 2;
					}
				} else {
					if (Time > (*Input->Processes)[i].TermTime) {
						if ((*Input->Processes)[i].Signalling_State == 0) {
							message(3, _("=> SIGTERM Processes[%d] [PID %d], t = %.0f, KillTime = %.0f\n"), i, (*Input->Processes)[i].PID, Time, (*Input->Processes)[i].KillTime);
							(*Input->Processes)[i].Signal = SIGTERM;
							kill((*Input->Processes)[i].PID, SIGTERM);
							(*Input->Processes)[i].Signalling_State = 1;
						}
					}
				}
			}
		}
		usleep(100*1000);	/* 100ms */
	}
	Input->Running = -1;
	pthread_exit(NULL);
	return NULL;
}

void* Process_Host_Thread(void *Parameters) {
	struct Process_Host_Thread_InputData *Input = (struct Process_Host_Thread_InputData*)Parameters;
	char Buffer[BUFSIZE_SHORT];
	char FileBuffer[BUFSIZE_SHORT];
	int Result, ProcID;
	int fd;

	strcpy(FileBuffer, TEMPPATH "sambascanner_process_hostXXXXXX");
	errno = 0;
	fd = (int)mkstemp(FileBuffer);
	if (errno != 0) {
		message(1, _("Couldn't create temporary file in Process_Host_Thread: %s\n"), strerror(errno));
		exit(EXIT_FAILURE);
	}

	message(2, _("Processing host: '%s'\n"), Input->Hostname);
	snprintf(Buffer, sizeof(Buffer), "-g -N -L //%s", Input->Hostname);
	ProcID = Exec_Timeout(Options.SmbClient_Executable, Buffer, TIMEOUT_PROCESSHOST_TERM, TIMEOUT_PROCESSHOST_KILL, fd, Input->Processes, Input->NumberProcesses);
	Result = WaitFor(ProcID, Input->Processes);

	pthread_mutex_lock(Input->NumberMutex);
	if (Result < 0) {
		/* Process was killed */
		(*Input->Hostlookup_Processes_Failed)++;
	} else {
		(*Input->Hostlookup_Processes_Succeeded)++;
	}
	pthread_mutex_unlock(Input->NumberMutex);

	Parse_Hostinfo(Input->Hostname, FileBuffer, Input->Hostlist, Input->Sharelist);
	if (!Input->PreserveTemp) unlink(FileBuffer);
	
	free(Input);
	ThreadCnt_Dec();
	pthread_exit(NULL);
	return NULL;
}

void* Process_Share_Thread(void *Parameters) {
	struct Process_Share_Thread_InputData *Input = (struct Process_Share_Thread_InputData*)Parameters;
	char CmdParameters[BUFSIZE_SHORT];
	char Buffer[BUFSIZE_SHORT];
	char Buffer2[BUFSIZE_SHORT];
	char FileBuffer[BUFSIZE_SHORT];
	int Result, ProcID;
	int fd;

	strcpy(FileBuffer, TEMPPATH "sambascanner_process_sharesXXXXXX");
	errno = 0;
	fd = (int)mkstemp(FileBuffer);
	if (errno != 0) {
		message(1, _("Couldn't create temporary file in Process_Share_Thread: %s\n"), strerror(errno));
		exit(EXIT_FAILURE);
	}
	Add_String_Atomic(Input->Resultfilelist, FileBuffer);
	
	message(2, _("Now revealing content of smb:%s with unique thread ID %d\n"), Input->Sharename, Input->UniqueThreadID);

	if (Options.TryPasswords) {
		CmdParameters[0] = 0;
	} else {
		strcpy(CmdParameters, "-g ");
	}
	if (Options.UserPass.Entries > 0) {
		int i;
		for (i = 0; i < Options.UserPass.Entries; i++) {
			snprintf(Buffer, sizeof(Buffer), "%s-u %s ", CmdParameters, Options.UserPass.Entry[i]);
			strcpy(CmdParameters, Buffer);
		}
	}

	Escape_Apostrophes(Buffer2, Input->Sharename, sizeof(Buffer2));
	snprintf(Buffer, sizeof(Buffer), "%s'smb:%s'", CmdParameters, Buffer2);
	strcpy(CmdParameters, Buffer);

	ProcID = Exec_Timeout(Options.SmbLister_Executable, CmdParameters, TIMEOUT_SMBLISTER_TERM, TIMEOUT_SMBLISTER_KILL, fd, Input->Processes, Input->NumberProcesses);
	Result = WaitFor(ProcID, Input->Processes);
	
	ThreadCnt_Dec();
	pthread_exit(NULL);
	return NULL;
}

void *Process_IP_Address_Thread(void *Parameters) {
	struct Process_IP_Address_Thread_InputData *Input = (struct Process_IP_Address_Thread_InputData*)Parameters;
	char Buffer[BUFSIZE_SHORT];
	char FileBuffer[BUFSIZE_SHORT];
	int Result, ProcID;
	int fd;

	message(3, _("Now processing IP %d.%d.%d.%d\n"), IP_ADDRESS(Input->IP_Address));

	strcpy(FileBuffer, TEMPPATH "sambascanner_process_ipaddrXXXXXX");
	errno = 0;
	fd = (int)mkstemp(FileBuffer);
	if (errno != 0) {
		message(1, _("Couldn't create temporary file in Process_IP_Address_Thread: %s\n"), strerror(errno));
		exit(EXIT_FAILURE);
	}
	
	snprintf(Buffer, sizeof(Buffer), "-A %d.%d.%d.%d", IP_ADDRESS(Input->IP_Address));
	ProcID = Exec_Timeout(Options.NmbLookup_Executable, Buffer, TIMEOUT_IPSCAN_TERM, TIMEOUT_IPSCAN_KILL, fd, Input->Processes, Input->NumberProcesses);
	Result = WaitFor(ProcID, Input->Processes);

	Parse_IP_Address_Info(FileBuffer, Input->Hostlist);
	if (!Input->PreserveTemp) unlink(FileBuffer);

	free(Parameters);
	ThreadCnt_Dec();
	pthread_exit(NULL);
	return NULL;
}

int pthread_create_detached(pthread_t *thread, void *(*start_routine)(void*), void *arg) {
	pthread_attr_t Attrs;
	pthread_attr_init(&Attrs);
	pthread_attr_setdetachstate(&Attrs, PTHREAD_CREATE_DETACHED);
	return pthread_create(thread, &Attrs, start_routine, arg);
}

