/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/wait.h>
#include <time.h>
#include <pthread.h>
#include <sys/time.h>
#include <libintl.h>
#include <locale.h>

#include "String.h"
#include "Mutex.h"
#include "Parser.h"
#include "CommandLine.h"
#include "Tools.h"
#include "IP_Range.h"
#include "Threads.h"
#include "Processes.h"
#include "PremiseCheck.h"
#include "sambascanner.h"

static struct Process *Processes = NULL;
static int NumberProcesses = 0;
struct Options_Structure Options;

void *ChildHandler_Thread(void *Parameter) {
	while (1) {
		int i;
		pid_t PID;
		Lock_Processes();
		for (i = 0; i < NumberProcesses; i++) {
			if (Processes[i].Valid == 1) {
				PID = waitpid(Processes[i].PID, &Processes[i].Status, WNOHANG);
				if (PID > 0) {
					/* Child really exited */
					Processes[i].Running = 0;
				}
			}
		}
		Unlock_Processes();
		sleep(1);
	}
	pthread_exit(NULL);
	return NULL;
}

void Internationalization() {
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
}

int main(int argc, char **argv) {
	double StartTime, EndTime;
	StringList Sharelist;
	StringList Hostlist;
	StringList Resultfilelist;
	static struct Process_Watcher_Thread_InputData Process_Watcher_Data;

	Internationalization();

	StartTime = GetTime();

	Init_StringList(&Sharelist);
	Init_StringList(&Hostlist);
	Init_StringList(&Resultfilelist);

	if (!Process_Parameters(argc, argv, &Hostlist))	exit(EXIT_FAILURE);
	if (!Check_Overall_Sanity()) exit(EXIT_FAILURE);

	Init_Semaphores(Options.ThreadCount);
	{
		pthread_t ProcessWatcher;
	
		Process_Watcher_Data.Running = 1;
		Process_Watcher_Data.NumberProcesses = &NumberProcesses;
		Process_Watcher_Data.Processes = &Processes;

		pthread_create_detached(&ProcessWatcher, Process_Watcher_Thread, &Process_Watcher_Data);
	}
	
	{
		pthread_t ChildHandler;
		pthread_create_detached(&ChildHandler, ChildHandler_Thread, &Process_Watcher_Data);
	}

	if (Options.Workgroup[0] == 0) {
		int ProcID;
		int Result;
		int fd;

		/* Workgroup was not set in options */
		message(3, _("STAGE: Now entering stage one.\n"));
		message(1, _("Looking up workgroup...\n"));

		fd = open(FILE_WORKGROUPINFO, O_WRONLY | O_TRUNC | O_CREAT, 0644);
		if (fd == -1) {
			message(1, _("Couldn't open file '%s' for writing: %s\n"), FILE_WORKGROUPINFO, strerror(errno));
			exit(EXIT_FAILURE);
		}
		ProcID = Exec_Timeout(Options.NmbLookup_Executable, "-M -S -- -", TIMEOUT_LOOKUPWG_TERM, TIMEOUT_LOOKUPWG_KILL, fd, &Processes, &NumberProcesses);
		Result = WaitFor(ProcID, &Processes);
		close(fd);
		if (Result != 0) {
			message(1, _("Process terminated at looking up workgroup.\n"));
			return 1;
		}
		if (Parse_Groupinfo(FILE_WORKGROUPINFO, Options.Workgroup)!=0) {
			message(1, _("Process terminated at parsing workgroup info list.\n"));
			return 2;
		};
		message(2, _("Returned workgroup: '%s'\n"), Options.Workgroup);
		if (!Options.PreserveTemp) unlink(FILE_WORKGROUPINFO);
		
		if (strlen(Options.Workgroup) == 0) {
			message(1, _("No workgroup found, aborting!\n"));
			message(1, _("Try setting the workgroup manually using the '-w' command line switch.\n"));
			return 3;
		}
	} else {
		message(3, _("STAGE: Stage one skipped.\n"));
	}
	
	if (Have_Patterns()) {
		int i;
		uint32_t IP;
		message(3, _("STAGE: Now entering stage two.\n"));
		message(1, _("Scanning IP addresses...\n"));
		ThreadCnt_Set(Options.ThreadCount);
		while (Get_Next_IP(&IP)) {
			pthread_t TempThread;
			struct Process_IP_Address_Thread_InputData *Data;
			char IP_String[16];
		
			IP_To_String(IP, IP_String, sizeof(IP_String));
			message(2, _("Scanning address %s\n"), IP_String);
			Data = (struct Process_IP_Address_Thread_InputData*)malloc(sizeof(struct Process_IP_Address_Thread_InputData));
			Data->IP_Address = IP;
			Data->Hostlist = &Hostlist;
			Data->UniqueThreadID = GetUniqueThreadID();
			Data->PreserveTemp = Options.PreserveTemp;
			Data->Processes = &Processes;
			Data->NumberProcesses = &NumberProcesses;

			ThreadCnt_Inc();
			message(4, _("Starting thread [UniqTID = %d], Process_IP_Address_Thread\n"), Data->UniqueThreadID);
			pthread_create_detached(&TempThread, Process_IP_Address_Thread, Data);
			message(4, _("Thread [UniqTID = %d] received pTID = %d\n"), Data->UniqueThreadID, (int)TempThread);
		}
		
		message(3, _("Waiting for threads to finish...\n"));
		for (i = 0; i < Options.ThreadCount; i++) ThreadCnt_Inc();	/* Wait until all threads are finished */
	} else {
		message(3, _("STAGE: Stage two skipped, no IP patterns were supplied.\n"));
	}
	
	{
		int ProcID;
		int Result;
		int fd;
		char Buffer[BUFSIZE_SHORT];
		message(3, _("STAGE: Now entering stage three.\n"));
		snprintf(Buffer, sizeof(Buffer), "-M -S %s", Options.Workgroup);
		message(1, _("Retrieving master browsers in workgroup '%s'...\n"), Options.Workgroup);
		fd = open(FILE_MASTERBROWSER, O_WRONLY | O_TRUNC | O_CREAT, 0644);
		if (fd == -1) {
			message(1, _("Couldn't open file '%s' for writing: %s\n"), FILE_MASTERBROWSER, strerror(errno));
			exit(EXIT_FAILURE);
		}
		ProcID = Exec_Timeout(Options.NmbLookup_Executable, Buffer, TIMEOUT_LOOKUPMB_TERM, TIMEOUT_LOOKUPMB_KILL, fd, &Processes, &NumberProcesses);
		Result = WaitFor(ProcID, &Processes);
		close(fd);
		if (Result != 0) {
			message(1, _("Process terminated at retrieving master browsers.\n"));
			return 1;
		}
		if (Parse_Masterbrowser(FILE_MASTERBROWSER, &Hostlist) != 0) {
			message(1, _("Process terminated at parsing master browser list.\n"));
			return 1;
		}
		if (!Options.PreserveTemp) unlink(FILE_MASTERBROWSER);
	}
	
	{
		int i;
		message(3, _("Number of hosts in list after stage three: %d\n"), Hostlist.Entries);
		for (i = 0; i < Hostlist.Entries; i++) {
			message(3, "%3d: %s\n", i + 1, Hostlist.Entry[i]);
		}
	}
	
	message(3, _("STAGE: Now entering stage four.\n"));
	message(1, _("Discovering hosts and share names...\n"));
	{
		int i;
		int NumberKilled, NumberOkay;
		int Hostlookup_Processes_Succeeded = 0;
		int Hostlookup_Processes_Failed = 0;
		pthread_mutex_t NumberMutex;

		pthread_mutex_init(&NumberMutex, NULL);

		NumberKilled = 0;
		NumberOkay = 0;
		ThreadCnt_Set(Options.ThreadCount);
		for (i = 0; i < Hostlist.Entries; i++) {
			/* Cannot allocate Process_Host_Data before loop statically as hostlist might grow during scan of hosts (host discovery) */
			struct Process_Host_Thread_InputData *Process_Host_Data;
			pthread_t TempThread;

			Process_Host_Data = (struct Process_Host_Thread_InputData*)malloc(sizeof(struct Process_Host_Thread_InputData));

			Process_Host_Data->Hostname = Hostlist.Entry[i];
			Process_Host_Data->Hostlist = &Hostlist;
			Process_Host_Data->Sharelist = &Sharelist;
			Process_Host_Data->NumberMutex = &NumberMutex;
			Process_Host_Data->Hostlookup_Processes_Succeeded = &Hostlookup_Processes_Succeeded;
			Process_Host_Data->Hostlookup_Processes_Failed = &Hostlookup_Processes_Failed;
			Process_Host_Data->PreserveTemp = Options.PreserveTemp;
			Process_Host_Data->Processes = &Processes;
			Process_Host_Data->NumberProcesses = &NumberProcesses;

			ThreadCnt_Inc();
			pthread_create_detached(&TempThread, Process_Host_Thread, Process_Host_Data);

			if (i == Hostlist.Entries - 1) {
				int j;
				/* Wait until threads are done as hosts might be added to list after threads are finished */
				for (j = 0; j < Options.ThreadCount; j++) ThreadCnt_Inc();
				ThreadCnt_Set(Options.ThreadCount);
			}
		}
		message(3, _("Waiting for threads to finish...\n"));
		for (i = 0; i < Options.ThreadCount; i++) ThreadCnt_Inc();	/* Wait until all threads are finished */
		message(2, _("%d processes killed (%.1f%%), %d returned sucessfully (%.1f%%).\n"), 
				Hostlookup_Processes_Failed, 
				100.0 * Hostlookup_Processes_Failed / (Hostlookup_Processes_Failed + Hostlookup_Processes_Succeeded), 
				Hostlookup_Processes_Succeeded, 
				100.0 * Hostlookup_Processes_Succeeded / (Hostlookup_Processes_Failed + Hostlookup_Processes_Succeeded)
			);
		message(2, _("Total hosts found: %d\n"), Hostlist.Entries);
	}
	
	{
		int i;
		message(3, _("Number of hosts in list after stage four: %d\n"), Hostlist.Entries);
		for (i = 0; i < Hostlist.Entries; i++) {
			message(3, "%3d: %s\n", i + 1, Hostlist.Entry[i]);
		}
	}
	

	/* Now try to reveal all share content on the found shares */
	message(3, _("STAGE: Now entering stage five.\n"));
	{
		int i;
		struct Process_Share_Thread_InputData *Process_Share_Data;
		
		ThreadCnt_Set(Options.ThreadCount);
		message(1, _("Revealing share content of %d %s in %d %s...\n"), Sharelist.Entries, SingularPlural(_("share"), _("shares"), Sharelist.Entries), Hostlist.Entries, SingularPlural(_("host"), _("hosts"), Sharelist.Entries));
		
		Process_Share_Data = (struct Process_Share_Thread_InputData*)malloc(sizeof(struct Process_Share_Thread_InputData) * Sharelist.Entries);
		for (i = 0; i < Sharelist.Entries; i++) {
			pthread_t TempThread;
			ThreadCnt_Inc();
			Process_Share_Data[i].Sharename = Sharelist.Entry[i];
			Process_Share_Data[i].UniqueThreadID = GetUniqueThreadID();
			Process_Share_Data[i].Resultfilelist = &Resultfilelist;
			Process_Share_Data[i].Processes = &Processes;
			Process_Share_Data[i].NumberProcesses = &NumberProcesses;
			message(4, _("Starting thread [%d], Process_Share_Thread\n"), Process_Share_Data[i].UniqueThreadID);
			pthread_create_detached(&TempThread, Process_Share_Thread, Process_Share_Data + i);
			message(4, _("Thread [UniqTID = %d] received pTID = %d\n"), Process_Share_Data[i].UniqueThreadID, (int)TempThread);
		}
	
		/* Now wait for the last few threads to come home */
		message(3, _("Waiting for threads to finish...\n"));
		for (i = 0; i < Options.ThreadCount; i++) ThreadCnt_Inc();	/* Wait until all threads are finished */
		message(3, _("...all done.\n"));
	
		/* Then, free memory */
		free(Process_Share_Data);
		Destroy_StringList(&Sharelist);
	}

	{
		int i;
		FILE *f, *g;
		char Line[BUFSIZE];
		int OutfileCnt = 0;
	
		f = fopen(Options.Outfile, "w");
		if (f == NULL) {
			message(1, _("Error opening outfile '%s' for writing (%s)\n"), Options.Outfile, strerror(errno));
		} else {
			for (i = 0; i < Resultfilelist.Entries; i++) {
				message(3, _("Merging file '%s'...\n"), Resultfilelist.Entry[i]);
				g = fopen(Resultfilelist.Entry[i], "r");
				if (g != NULL) {
					while (fgets(Line, BUFSIZE, g) != NULL) {
						if (strlen(Line) > 0) {
							if (Line[strlen(Line) - 1] == '\n') Line[strlen(Line) - 1] = '\0';
							if (strncmp(Line, "smb://", 6) == 0) {
								fprintf(f, "%s\n", Line);
								OutfileCnt++;
							}
						}
					}
					fclose(g);
				}
				if (!Options.PreserveTemp) unlink(Resultfilelist.Entry[i]);
			}
			fclose(f);
		}
		message(2, _("Found a total of %d files in the network.\n"), OutfileCnt);
		if (OutfileCnt == 0) {
			message(1, _("No files were found during the scanning process. Try the following to improve\nsearch results:\n"));
			message(1, _("    * Use the '-t' command line switch (e.g. -t 10)\n"));
			message(1, _("    * Use the '-i' command line switch (e.g. -i 192.168.1.0/24)\n"));
		}
	}

	/* Wait for process watcher to die */
	message(3, _("STAGE: Now entering stage six.\n"));
	
	/* Tell process watcher to exit */
	Process_Watcher_Data.Running = 0;

	/* And wait for its acknowledgement */
	while (Process_Watcher_Data.Running != -1) {
		usleep(100 * 1000);
	}

	Destroy_StringList(&Hostlist);
	Destroy_StringList(&Resultfilelist);
	Free_All_Patterns();

	EndTime = GetTime();
	message(1, _("Complete scanning process took: %d:%02d min/sec\n"), ((int)(EndTime - StartTime)) / 60, ((int)(EndTime - StartTime)) % 60);
	return 0;
}

